<?php

namespace Drupal\term_glossary\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * The TermGlossary controller.
 */
class TermGlossaryController extends ControllerBase {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Drupal\Core\Config\ConfigManagerInterface definition.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected $configManager;

  /**
   * Symfony\Component\HttpFoundation\RequestStack definition.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The entity repository service.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);
    $instance->database = $container->get('database');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->configManager = $container->get('config.manager');
    $instance->requestStack = $container->get('request_stack');
    $instance->entityRepository = $container->get('entity.repository');
    $instance->moduleHandler = $container->get('module_handler');
    return $instance;
  }

  /**
   * Api Search Per Letter.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   Json response.
   */
  public function apiSearchPerLetter($letter) {
    $results = ['message' => $this->t('No results found for @letter', ['@letter' => $letter])];
    $status = 200;
    if (!empty($letter) && preg_match("/^[a-zA-Z]$/", strtoupper($letter))) {
      $query = $this->entityTypeManager->getStorage('taxonomy_term')->getQuery();
      $vid = $this->getTheCorrectVocab();
      $query->condition('name', $letter, 'STARTS_WITH');
      $query->condition('vid', $vid);
      $query->condition('langcode', $this->languageManager()->getCurrentLanguage()->getId());
      $query->accessCheck();
      $entity_ids = $query->execute();
      if (count($entity_ids) !== 0) {
        $results = [];
        $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
        $terms = $term_storage->loadMultiple($entity_ids);
        foreach ($terms as $term) {
          $term = $this->entityRepository->getTranslationFromContext($term);
          $results[] = [
            'name' => $term->getName(),
            'tid' => $term->id(),
            'description' => $term->getDescription(),
          ];
        }
        $this->moduleHandler->invokeAll('term_glossary_alter_results', [
          &$results,
          $terms,
          $letter,
        ]);
      }
    }
    return new JsonResponse($results, $status);
  }

  /**
   * Helper to get term by id.
   */
  public function apiGetTermById($tid) {
    $results = ['message' => $this->t('No results found for @tid', ['@tid' => $tid])];
    $status = 200;
    $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
    $term = $term_storage->load($tid);
    if (!empty($term)) {
      $term = $this->entityRepository->getTranslationFromContext($term);
      $results = [
        'name' => $term->getName(),
        'tid' => $term->id(),
        'description' => $term->getDescription(),
      ];
    }
    $this->moduleHandler->invokeAll('term_glossary_alter_result', [
      &$results,
      $term,
      $tid,
    ]);
    return new JsonResponse($results, $status);
  }

  /**
   * Helper to search per term.
   */
  public function apiSearchPerTerm() {
    $request = $this->requestStack->getCurrentRequest();
    // 200 because js is not checking for error.
    $status = 200;
    $results = ['message' => $this->t('error bad request')];
    if (!empty($request->get('t'))) {
      // @todo make sure this is "safe" here.
      $term = Html::escape(trim($request->get('t')));
      if (!empty($term)) {
        $results = [];
        $term_storage = $this->entityTypeManager->getStorage('taxonomy_term');
        $query = $term_storage->getQuery();
        $vid = $this->getTheCorrectVocab();
        $query->condition('name', $term, 'CONTAINS');
        $query->condition('vid', $vid);
        $query->condition('langcode', $this->languageManager()->getCurrentLanguage()->getId());
        $query->accessCheck();
        $entity_ids = $query->execute();

        if (count($entity_ids) == 0) {
          $query = $term_storage->getQuery();
          $vid = $this->getTheCorrectVocab();
          $query->condition('name', "%" . $this->database->escapeLike($term) . "%", 'LIKE');
          $query->condition('vid', $vid);
          $query->condition('langcode', $this->languageManager()->getCurrentLanguage()->getId());
          $query->accessCheck();
          $entity_ids = $query->execute();
        }
        if (count($entity_ids) == 0) {
          $results = ['message' => $this->t('No results found for search term "@term"', ['@term' => $term])];
        }
        else {
          $terms = $term_storage->loadMultiple($entity_ids);
          foreach ($terms as $term) {
            $term = $this->entityRepository->getTranslationFromContext($term);
            $results[$term->id()] = [
              'name' => $term->getName(),
              'tid' => $term->id(),
              'description' => $term->getDescription(),
            ];
            // Perhaps fire hook to include more.
          }
          $this->moduleHandler->invokeAll('term_glossary_alter_results', [
            &$results,
            $terms,
            $term,
          ]);
        }

      }
    }
    return new JsonResponse($results, $status);
  }

  /**
   * Function to get the vocab name form config.
   *
   * @return string
   *   the vocab name from config.
   */
  private function getTheCorrectVocab() {
    // @todo replace this with config.
    $config = $this->configManager->getConfigFactory();
    $config = $config->get('term_glossary.glossaryconfig');
    if (empty($config->get('vocab'))) {
      throw new \Exception('glossary has not been configured yet');
    }
    return $config->get('vocab');
  }

}
