<?php

namespace Drupal\term_glossary\Service;

use Drupal\Component\Utility\Html;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigManagerInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\taxonomy\TermInterface;

/**
 * The TermGlossary manager service.
 */
class TermGlossaryManager {

  /**
   * The config manager.
   *
   * @var \Drupal\Core\Config\ConfigManagerInterface
   */
  protected ConfigManagerInterface $configManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The entity term storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $termStorage;

  /**
   * The cache service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected CacheBackendInterface $cache;

  /**
   * The configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The cache id.
   *
   * @var string
   */
  protected string $cacheId = 'glossary_terms_array';

  /**
   * The seen terms.
   *
   * @var array
   */
  protected static array $seenTerms = [];

  public function __construct(ConfigManagerInterface $configManager, EntityTypeManagerInterface $entityTypeManager, CacheBackendInterface $cache) {
    $this->configManager = $configManager;
    $this->entityTypeManager = $entityTypeManager;
    $this->cache = $cache;
    $this->termStorage = $this->entityTypeManager->getStorage('taxonomy_term');
    $this->config = $configManager->getConfigFactory()->get('term_glossary.glossaryconfig');
  }

  /**
   * Inject terms in the field value.
   *
   * @param string $input
   *   Original field string.
   * @param string $vocabulary
   *   The vocabulary name to use.
   *
   * @return string
   *   String with highlighted terms.
   */
  public function replaceFieldValue($input, $vocabulary = NULL) {
    if (!empty($input)) {

      static::$seenTerms = [];

      $termList = $vocabulary ? $this->getTerms($vocabulary) : $this->getTerms();
      $is_single__match = $this->config->get('single_match');
      $is_full_word = $this->config->get('full_word');

      $html_dom = Html::load($input);
      $xpath = new \DOMXPath($html_dom);

      /*
       * Parts and ideas of the code have been taken from the "glossify" module.
       * https://git.drupalcode.org/project/glossify
       */
      /** @var \DOMNodeList $textNodes */
      $textNodes = $xpath->query('//text()[not(ancestor::a) and not(ancestor::img)]');
      foreach ($textNodes as $originalNode) {
        $text = $originalNode->nodeValue;
        // ->nodeValue turns escaped entities back and that created "invalid"
        // fragments, this escapes it back
        $text = Html::escape($text);

        if (empty(trim($text))) {
          continue;
        }

        $new_value = $text;
        $parentNode = $originalNode->parentNode;
        $nextNode = $originalNode->nextSibling;
        $parentNode->removeChild($originalNode);

        $replacements = [];
        $replacement_index = 0;

        foreach ($termList as $term_id => $term_array) {
          $name = preg_quote($term_array['name'], '/');
          $pattern = $is_full_word ? "/(\b$name\b)+/i" : "/$name+/i";
          preg_match_all($pattern, $new_value, $matches);
          if (is_array($matches[0]) && count($matches[0]) >= 1) {
            $matches = array_unique($matches[0]);

            foreach ($matches as $match) {
              if ($is_single__match) {
                if (in_array(mb_strtolower($match), static::$seenTerms)) {
                  continue;
                }
                static::$seenTerms[] = mb_strtolower($match);
              }
              $replacements[$replacement_index] = '<span role="button" tabindex="0" class="glos-term" data-gterm="' . $term_id . '">' . $match . '</span>';
              $offset = strpos($new_value, $match);
              $length = strlen($match);
              $new_value = substr_replace($new_value, '{{' . $replacement_index . '}}', $offset, $length);
              $replacement_index += 1;
            }
          }
        }

        // Regular expression pattern to match {{number}}.
        $replacement_pattern = '/\{\{(\d+)\}\}/';
        // Replace callback function.
        $new_value = preg_replace_callback($replacement_pattern, function ($matches) use ($replacements) {
          // Get the number inside the {{ }}.
          $index = (int) $matches[1];
          // Return the corresponding item from the array.
          return $replacements[$index] ?? '{{' . $index . '}}';
        }, $new_value);

        $frag = $html_dom->createDocumentFragment();
        $frag->appendXML($new_value);

        $parentNode->insertBefore($frag, $nextNode);
      }

      return Html::serialize($html_dom);
    }
    else {
      return $input;
    }

  }

  /**
   * Get configuration.
   *
   * @param string $key
   *   Config key.
   *
   * @return string
   *   Config value.
   */
  public function getConfigValue(string $key) :string {
    return $this->config->get($key);
  }

  /**
   * Get an updated list of glossary terms.
   *
   * @return array
   *   Glossary term list
   */
  private function getTerms($vocabulary = NULL) {
    $vocab = $vocabulary ?: $this->config->get('vocab');
    $cacheId = $this->cacheId . '-' . $vocab;
    $data = $this->cache->get($cacheId);
    if (empty($data)) {
      $terms = $this->updateTermList($vocab);
    }
    else {
      $query = $this->termStorage->getQuery();
      $query->condition('vid', $vocab);
      $query->accessCheck();
      $entity_ids = $query->execute();
      // Someone could have added new term since last cache.
      if (count($entity_ids) == count($data->data)) {
        // Get the list from cache.
        $terms = $data->data;
      }
      else {
        // Set cache again.
        $terms = $this->updateTermList($vocab);
      }

    }
    return $terms;
  }

  /**
   * Update term list.
   *
   * @param string $vocab
   *   Glossary vocabulary id.
   *
   * @return array
   *   Glossary term list
   */
  private function updateTermList($vocab) {
    $terms = $this->termStorage->loadByProperties(['vid' => $vocab]);
    // Sort terms by weight to allow processing prioritization.
    usort($terms, function (TermInterface $a, TermInterface $b) {
      return $a->getWeight() <=> $b->getWeight();
    });
    $terms_array = [];
    /**
     * @var  integer $tid
     * @var \Drupal\taxonomy\Entity\Term $term
     */
    foreach ($terms as $tid => $term) {
      $terms_array[$tid] = [
        'tid' => $tid,
        'name' => $term->getName(),
      ];
    }
    $cacheId = $this->cacheId . '-' . $vocab;
    $this->cache->set($cacheId, $terms_array, CacheBackendInterface::CACHE_PERMANENT);
    return $terms_array;
  }

}
